/* ---------------------------------------------------------------------*/
/*  feature_extraction.c       Teemu Kilpelinen 17.05.2001				*/
/*   												          	        */
/*  This module includes FeatureExtraction -function. It is the			*/
/*	most important function in the speech2txt -program.                 */                                                               
/*	Version Number: 0.01											   */
/* ---------------------------------------------------------------------*/

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "feature_extraction.h"
#include "fourier.h"
#include "frame.h"
#include "helpfunctions.h"
#include "mel-functions.h"
#include "memctrl.h"
#include "s2t_defs.h"


/* --------------------------------------------------------------- */
FEATURE_TABLE FeatureExtraction(SND_BUFFER signal, FLOAT_ARRAY window_function, 
	USER_PARAMETERS* user_params, SIGNAL_INFORMATION* signal_variables, 
	FRAME_INFORMATION* frame_variables, int* count_of_result_vectors){
/*	
	Functioning:	
	
	Inputs:			SND_BUFFER signal					 Signal.
					FLOAT_ARRAY window_function			 Values of window function.			
					USER_PARAMETERS* user_params	 	 struct of user variables.
					SIGNAL_INFORMATION* signal_variables struct of signal variables.
					FRAME_INFORMATION* frame_variables	 struct of frame variables.
					int* count_of_result_vectors		 Count of vectors in cepstrum. 
		
	Outputs:		-
	Return value:	Cepstrum array.
	History:		Teemu Kilpelinen 18.05.2000 
	Version:		1.0 (original)

	Reference:		-
*/
	int		
		    /* index variable */
			vector_element,
		 
			/* count of vectors in cepstrum */
			count_of_vectors;
				
	FFT_FRAME
			
			/* Struct in which are variables needed in fft_float -function */ 
			fft_frm;
		
	FLOAT_ARRAY
		 
			/* An Array where are the values of the mel-filter bank */
	        mel_filter_bank = NULL,
		
			/* An Array where are the mel-filtered values of frame */
			log_filtered_spectrum,

			/* An Array where are the values of liftering function. 
			   Length of array == user_params->MFCC_ord */
			lifter_function;
			
	FRAME_ARRAY 
		  
		    /* Includes short part of signal */
			frame;
						
	FILTER_BANK
			
			/* mel filter bank */
			*filter_bank;

	FEATURE_TABLE 
			
			/* The result array of DCT-function. Count of rows == 
			   count of vectors. Count of columns == user_params->MFCC-ord. 
			*/
			cepstrum;	

	/*Calculates size of the frame (which is power of two) */
	frame_variables->length_pow_of_two = NearestPowerOfTwo
		(frame_variables->samples_per_frame, 
		user_params->use_nearest_small_pow_of_two);
	
	/*Calculates how long the window shifting is in samples */
	frame_variables->win_shift_samples = Ms2Samples(user_params->win_shift_ms, 
		signal_variables->sampling_rate);   
	
	if (user_params->num_mel_filters == 0)
		user_params->num_mel_filters = 
		NumberOfMelFilters(signal_variables->sampling_rate);

	/* Design filter bank */ 
	filter_bank = CreateFilterBank(user_params->num_mel_filters, 
		frame_variables->length_pow_of_two, 
		signal_variables->sampling_rate);
	
	/* Calculates values to window_fuction (depends of user choices: Hamming, 
	   Blackman, etc.) */
	lifter_function = CreateNewFloatArray(user_params->MFCC_ord);
	LifteringCoefficients(lifter_function, user_params->lifter_type, 
		user_params->MFCC_ord);

	count_of_vectors = (int)(signal_variables->count_of_samples/
		(float)frame_variables->win_shift_samples);

	/* Memory allocations for frames. Calloc makes a frame filled with 
	   zeroes. */
	frame = calloc(sizeof(float) * frame_variables->length_pow_of_two, 
			sizeof(float));
	fft_frm.real_part = CreateNewFloatArray(frame_variables->length_pow_of_two);
	fft_frm.im_part = CreateNewFloatArray(frame_variables->length_pow_of_two);
	memset(frame, 0, sizeof(float) * frame_variables->length_pow_of_two); 
	
	log_filtered_spectrum = CreateNewFloatArray(user_params->num_mel_filters);  
	
	/* Memory allocation for result vector-array. +1 for Include0thCoeff!!!. */
	cepstrum = CreateNewFT(count_of_vectors, user_params->MFCC_ord + 1);	
	
	/* Set the index, where from the frame begin  */
	frame_variables->begin = 0;

	count_of_vectors = 0;

	/* Loop where is implemented the handling of frames and FFT itself */
	while (frame_variables->begin + frame_variables->samples_per_frame < 
		signal_variables->count_of_samples){
		
	count_of_vectors++;	

		MakeFrame(signal, frame_variables->begin, 
			frame_variables->samples_per_frame, 
			frame_variables->length_pow_of_two, frame, window_function, 
			user_params->use_nearest_small_pow_of_two);
	
		fft_float(frame_variables->length_pow_of_two, 0, frame, 0, 	
			fft_frm.real_part, fft_frm.im_part);
		
		for (vector_element = 0; vector_element < 
			frame_variables->length_pow_of_two; vector_element++)
				frame[vector_element] = 
				ComplexAbs((float)fft_frm.real_part[vector_element], 
				(float)fft_frm.im_part[vector_element] + FLT_EPSILON);
		/* FLT_EPSILON added because log-filtering goes wrong, if there is
		   zero values in frames. FLT_EPSILON is the smallest possible decimal 
		   in float number */

		ComputeLogFilteredSpectrum(frame, log_filtered_spectrum, 
			filter_bank, user_params);
		
		DCT(cepstrum, log_filtered_spectrum, user_params->MFCC_ord, 
			user_params->num_mel_filters, count_of_vectors); 						
	
		/* Liftering if lifter_type != 1 (not truncated). */ 
		if (user_params->lifter_type != 1)
			CepstralLiftering(cepstrum[count_of_vectors - 1], lifter_function, 
				user_params->MFCC_ord);
			
		frame_variables->begin += frame_variables->win_shift_samples;
	}
	
	/* How many vectors managed to record in cepstrum */
	*count_of_result_vectors = count_of_vectors;

	/* memory deallocations */
	deallocate(frame);
	deallocate(fft_frm.real_part);
	deallocate(fft_frm.im_part);
	deallocate(log_filtered_spectrum);
	deallocate(mel_filter_bank);
	deallocate(lifter_function);
	return cepstrum;
} /* FeatureExtraction */	