/* --------------------------------------------------------------------*/
/*  helpfunctions.c       Teemu Kilpelinen 20.10.2000				   */
/*   												          	       */
/*  This module includes help functions. Here is functions for type    */
/*  conversions	etc..												   */
/*	Version Number: 0.01											   */
/* --------------------------------------------------------------------*/

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
//#include <unistd.h>
#include "helpfunctions.h"
#include "s2t_defs.h"

/* Macro which returns the bigger of two parameters  */
#define MAX(a,b) (a > b ? a : b) 


/* --------------------------------------------------------------- */
int Int16ToFloat(INT16* integer_buffer, SND_BUFFER signal, 
	int count_of_samples)
/*	
	Functioning:	Changes integer signal to float signal. 
	Inputs:			
					INT16* integer_buffer   source array (int*)
					SND_BUFFER signal           result array (float*)
					int count_of_samples    count of samples in source array						
	
	Outputs:		Signal in float* array.

	Return value:	-
	History:		Teemu Kilpelinen 15.09.2000.
	Version:		1.0 (original)

	Reference:		
*/
{	
	INT32 i;
	INT16 b;
	float c;
	
	for (i=0; i < count_of_samples; i++){
		b = integer_buffer[i];
        c = ((float) b ); /* /pow(2,16) */
		signal[i] = c;
		}
		
return 0;
} /* Int16ToFloat */


/* --------------------------------------------------------------- */
int Ms2Samples(int ms, int sampling_rate){
/*  
	Functioning:	Changes ms to samples.
	Inputs:			int ms		       Sample's length in milliseconds
					int sampling_rate  Sampling rate of input signal
	Outputs:		-
	Return value:	Sample's lenght in samples  
	History:		Teemu Kilpelinen and Tomi Kinnunen	 15.10.2000
	Version:		1.0 (original)

	Reference:		-

*/	  
	return (int)((ms/1000.0) * sampling_rate);
} /* Ms2Samples */

  
int NearestPowerOfTwo(int frame_length, int round_downwards){
/*	int	NearestPowerOfTwo(int frame_length, int round_downwards)
	
	Functioning:	Returns either nearest bigger or nearest smaller 
					power of two of frame_length. If round_downwards == 1,
					returns the nearest power of two <= frame_length. 
					Otherwise returns nearest power of two >= frame_length.  
	
	Inputs:			int frame_length	 Frame length in samples.
					int round_downwards  Boolean variable (1 or 0).
	Outputs:		-
	Return value:	Nearest either bigger or smaller power of two of frame_length  
	History:		Teemu Kilpelinen ja Ismo Krkkinen 20.12.2000 
	Version:		1.0 (original)

	Reference:		-
*/
    int power = 1;
    while (frame_length > power) power *= 2;
	if ((round_downwards)&&(frame_length < power)) power /= 2;

	return power;
} /* NearestPowerOfTwo */


/* --------------------------------------------------------------- */
float ComplexAbs(float real_part, float im_part){ 
/*	
	Functioning:	Returns the complex absolute value 
					of an imaginary number.
	Inputs:			float real_part		The real part of the 
										imaginary number.
					float im_part		The imaginary part of the 
										imaginary number.
	Outputs:		-
	Return value:	The complex absolute value 
	History:		Teemu Kilpelinen and Tomi Kinnunen		22.11.2000
	Version:		1.0 (original)

	Reference:		MAOL-taulukot (1995)
*/	  
	return (float)((double)sqrt((double)(real_part * real_part + 
		im_part * im_part)));
} /* ComplexAbs */


/* --------------------------------------------------------------- */
int Hz2Index(float frequency, int sampling_rate, int frame_length){
/*  
	Functioning:	Changes Hz to mel-scale.
	Inputs:			float frequency        In Hz.
					int sampling_rate      In Hz.
				    int frame_length       frame length in samples (power of two).	
					(WITHOUT decimation by 2 == window with added zeroes).	
	Outputs:		
	Return value:	index of the frequency in the FFT table.  
	History:		Teemu Kilpelinen      20.12.2000, 17.3.2001
	Version:		1.0 (original)

	Reference:		Discrete-Time Signal Processing of Speech Signals, 1999. 
					IEEE Press.
*/	  
    float Nyquist_freq, /* Nyquist-frequency of the signal */
	      freq_step;    /* Frequency step */
	int rounded_index;  /* round of nearest index */

    Nyquist_freq = sampling_rate/2.0f;  
	
	freq_step = Nyquist_freq/(frame_length/2.0f); 
		
	/* Round of nearest index */	
	rounded_index = (int)(frequency / freq_step);           
    
	/* if rounded_index is over frame_length, returns frame_length */
	if (rounded_index < frame_length/2.0)
		return rounded_index;
	else 
		return -1; 
} /* Hz2Index */


/* --------------------------------------------------------------- */
float Hz2Mel(float Hz_frequency){
/*
	Functioning:	f(mel) = 2595 * log10(1+ f/700)
	Inputs:			float Hz_frequency     Frequency in hertz.  
	
	Outputs:		
	Return value:	 
	History:		Teemu Kilpelinen and Tomi Kinnunen		22.01.2001
	Version:		1.0 (original)

	Reference: Discrete-Time Signal Processing of Speech Signals, 1999. 
			   IEEE Press.		
*/
	
	return 2595.0f * (float)log10(1 + Hz_frequency/700.0);
} /* Hz2Mel */


/* --------------------------------------------------------------- */
void LifteringCoefficients(FLOAT_ARRAY lifter_function, int lifter_type, 
	int lifter_length){
/*
	Functioning:	Compute coefficients for...  
	
	Inputs:			FLOAT_ARRAY liftering_function   Values of liftering function.
					int lifter_type                  1 = truncated,
													 2 = raised sine,
													 3 = half raised sine.
				    int lifter_length				 Length of liftering array.	
	Outputs:		Liftering function.
	
	Return value:	-
	History:		Teemu Kilpelinen 28.05.2001.
	Version:		1.0 (original)

	Reference:		Bin Zhen, Xihong Wu, Zhimin Liu and Huisheng Chi:
					"On the Use of Bandpass Liftering in Speaker Recognition",
					National Key Lab on Machine Perception, Center for 
					Information Science,
					Peking University, Beijing, 100871, China.
*/	  
  int k;
  float h;	
		
	/* truncated (== no liftering) */ 
	if (lifter_type == 1)
		return;

	/* raised sine */	
	if (lifter_type == 2){	
		h = lifter_length / 2.0f;
		for (k = 1; k <= lifter_length; k++)
			lifter_function[k - 1] = (float)(1 + (h * 
			sin((k * PI)/lifter_length))); 
	}
	
	/* half raised sine */
	else{ 
		for (k = 1; k <= lifter_length; k++)
			lifter_function[k - 1] = (float)(1 + (lifter_length/2 * 
			sin(1/2.0 * ((k * PI)/lifter_length))));
		}
}/* LifteringCoefficients */


/* --------------------------------------------------------------- */
void CepstralLiftering(FEATURE_VECTOR vector, FLOAT_ARRAY lifter_function, 
	int lifter_length){
/*  
	Functioning:	Multibly vector by lifter-function.
	Inputs:			FEATURE_ARRAY vector		vector 
					FLOAT_ARRAY lifter_function	liftering array
					int lifter_length			length of liftering function
	
	Outputs:		Liftered cepstral vector.
	Return value:	-  
	History:		Teemu Kilpelinen      23.05.2001
	Version:		1.0 (original)

	Reference:		-
*/	  
 int k;
	
 for (k = 1; k <= lifter_length; k++)
	 vector[k] = vector[k] * lifter_function[k - 1];
} /* CepstralLiftering */


/* --------------------------------------------------------------- */
void WriteFeatureTableToFile(FEATURE_TABLE table, int count_of_vectors, 
	int begin_element, int end_element, char* filename){
/*
    Functioning:    Returns text-file filled by cepstrum.	
	Inputs:			FEATURE_TABLE table,	Feature table to be written into file
					int count_of_vectors,	How many vectors
					int begin_element,		From which vector element to start ...
					int end_element,		... and where to stop.
					char* filename			Output file name.
		    
	Outputs:		Cepstrum in output file.
	Return value:	-
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/
	int		i, j;
	FILE	*filep;

	filep = fopen(filename,"w"); /* Open text file for writing */
	for (i = 0; i < count_of_vectors; i++){
		for(j = begin_element; j <= end_element; j++)			
			fprintf(filep,"%f ", table[i][j]);
		fprintf(filep,"\n");
	}
	fclose(filep); 
} /* WriteFeatureTableToFile */