/* --------------------------------------------------------------------*/
/*  mel-functions.c       Teemu Kilpelinen 22.5.2001				   */
/*                                                                     */
/*  This module includes functions for counting of mel.				   */
/*	Version Number: 0.01											   */
/* --------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "helpfunctions.h"
#include "mel-functions.h"
#include "memctrl.h"
#include "s2t_defs.h"


/* --------------------------------------------------------------- */
static FLOAT_ARRAY FillCoeffTable(int begin, int center, int end){
/*   
    Functioning:    Returns pointer to array filled by mel-coeffs.	
	Inputs:		    int begin	Index where from to begin 
					int center	Index of center point of filter
					int end	    Index where to stop
	Outputs:		-
	Return value:	Pointer to array filled by mel-coeffs.
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/	
	int i;
	FLOAT_ARRAY coeffs;
	
	coeffs = CreateNewFloatArray(end - begin + 1);
		
	for (i = 1; i <= center - begin; i++)
		coeffs[i - 1] = i/(float)(center - begin);
		
			for (i = 0; i <= end - center; i++)
		coeffs[center - begin + i] = (end - center - i)/(float)(end - center + 1);

	return coeffs;
} /* FillCoeffTable */			

		
/* --------------------------------------------------------------- */
FILTER_BANK* CreateFilterBank(int count_of_filters, int length /*pow of two */, 
	int sampling_rate){
/*   
    Functioning:    Returns array filled by mel-coeffs.	
	Inputs:		    int count_of_filters 
					int length (pow of two)  
					int sampling_rate
	Outputs:		-
	Return value:	Filter bank.
	History:		Teemu Kilpelinen ja Ismo Krkkinen 22.05.2001 
	Version:		1.0 (original)
*/	
	int k, prev_center = 0, next_center, curr_center;
	float mel_max, mel_freq_Hz, fmel;
	FILTER_BANK* filter_bank;

	mel_max = Hz2Mel(sampling_rate/2.0f);   

	filter_bank = allocate(sizeof(FILTER_BANK));

	filter_bank->num_mel_filters = count_of_filters;
    filter_bank->filters = allocate(count_of_filters * sizeof(FILTER));

	fmel = 1/((float)count_of_filters) * mel_max;
	mel_freq_Hz = (float)(700 * (pow(10,(fmel/2595.0)) - 1));
	curr_center = Hz2Index(mel_freq_Hz,		 sampling_rate, length);

	for (k = 1; k <= count_of_filters; k++){
		fmel = (k + 1)/((float)count_of_filters) * mel_max;
		mel_freq_Hz = (float)(700 * (pow(10,(fmel/2595.0)) - 1));
		
		if (k != count_of_filters)
			next_center = Hz2Index(mel_freq_Hz, sampling_rate, length);
		else
			next_center = curr_center;

		filter_bank->filters[k - 1].filter_begin = prev_center;
		filter_bank->filters[k - 1].filter_end   = next_center;
		filter_bank->filters[k - 1].coeffs       = FillCoeffTable(prev_center, 
			curr_center, next_center);
		/* Fill coefficient array */
		prev_center = curr_center;
		curr_center = next_center;
	}
	
	filter_bank->filters[count_of_filters - 1].filter_end = (int)(length/2.0 - 1);
	return filter_bank;
} /* CreateFilterBank */


/* --------------------------------------------------------------- */
FILTER* GetFilter(FILTER_BANK* filter_bank, int filter_index){
/*	
    Functioning:    Returns array filled by mel-coeffs.	
	Inputs:		    FILTER_BANK* filter_bank	Filter bank.
					int filter_index			Index in filter bank.
	Outputs:		-
	Return value:	One filter from filter_bank.
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/
	return &(filter_bank->filters[filter_index - 1]);
} /* GetFilter */ 


float FilterOutput(FILTER* filter, FLOAT_ARRAY frame){
/* 
	Functioning:    Filters frame with one filter.	
	Inputs:		    FILTER* filter		Pointer to one filter.
					FLOAT_ARRAY frame   Frame (array where the piece 
									    of signal have been copied).  
	Outputs:		
	Return value:	-
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/
	int k;
	float sum = 0;

	for (k = filter->filter_begin; k <= filter->filter_end; k++){
		sum = sum + (frame[k] * filter->coeffs[k - filter->filter_begin]);
	}

	/* the average of all result values. */
	return sum/(filter->filter_end - filter->filter_begin + 1);
} /* FilterOutput */


/* --------------------------------------------------------------- */
int NumberOfMelFilters(int sampling_rate){
/* 
    Functioning:    Returns the number of Mel-Filters 	
	Inputs:		    int sampling_rate  Sampling rate of input signal.	    
	Outputs:		-
	Return value:	The number of Mel-Filters
	History:		Teemu Kilpelinen 01.03.2001
	Version:		1.0 (original)
*/
	return (int)ceil((double)(3 * log(sampling_rate)));
} /* NumberOfMelFilters */


/* --------------------------------------------------------------- */
void ComputeLogFilteredSpectrum(FRAME_ARRAY frame, 
	FLOAT_ARRAY log_filtered_spectrum, FILTER_BANK* filter_bank,
	USER_PARAMETERS* user_params){ 
/*
	Functioning:    Computes the log-filtered spectrum of spectrum (fft's result)
	Inputs:			FRAME_ARRAY frame
					FLOAT_ARRAY log_filtered_spectrum
					FILTER_BANK* filter_bank
					USER_PARAMETERS user_params		User parameters
    Outputs:		An array which includes the values of log-filtered frame.

	Return value:	- 
	History:		Teemu Kilpelinen 11.04.2001
	Version:		1.0 (original)

	Reference:		Discrete-Time Processing of Speech Signals, 
					John R. Deller Jr., John H.L. Hansen, 
					John G. Proakis, 2000.  
*/
	int i;  

	/* Absolute magnitude domain */
	if (user_params->absolute_magnitude == 1)
		for (i = 1; i <= filter_bank->num_mel_filters; i++){
			log_filtered_spectrum[i - 1] = 
			(float)log10(FilterOutput(GetFilter(filter_bank, i), frame));
		}
	
	else /* Power domain */
		for (i = 1; i <= filter_bank->num_mel_filters; i++){
			log_filtered_spectrum[i - 1] = 
			(float)log10(pow(FilterOutput(GetFilter(filter_bank, i), frame), 2));
	}
}/* ComputeLogFilteredSpectrum */


/* --------------------------------------------------------------- */
void DCT(FEATURE_TABLE cepstrum,  FLOAT_ARRAY log_filtered_spectrum, 
		 int MFCC_ord, int num_mel_filters, int row){
/*	
	Functioning: Computes discrete cosine transform of log filtered spectrum.
	Inputs:			
		FEATURE_TABLE cepstrum		    An 2D array which will be filled by 
									    results of function.
	    FLOAT_ARRAY log_filtered_spectrum   The values of log-filtered 
											(fft's result) spectrum.
		int MFCC_ord					Number of cepstral coefficients. 
		int num_mel_filters				Number of mel-filters.
		int row						The current number of vector.
						
	Outputs:	 An DCT result-array.

	Return value:	-
	History:		Teemu Kilpelinen 26.04.2001
	Version:		1.0 (original)

	Reference:		
*/
	int n, k; 
	float sum;
	
	for	(n = 0; n <= MFCC_ord; n++){ 
		sum = 0.0;
		for (k = 1; k <= num_mel_filters; k++){
			sum = sum + (float)(log_filtered_spectrum[k - 1] * 
				cos(n * (k - 0.5) * (PI / num_mel_filters)));
		}
		cepstrum[row - 1][n] = sum;
	}

}/* DCT */	