/* --------------------------------------------------------------------*/
/*  pre_process.cpp       Teemu Kilpelinen 15.3.2001                  */
/*   												          	       */
/*	This module includes functions for signal pre-processing.		   */ 
/*	Version Number: 0.01											   */
/* --------------------------------------------------------------------*/

#include <stdio.h>
#include <windows.h>
#include <math.h>
#include "helpfunctions.h"
#include "memctrl.h"
#include "preprocess.h"
#include "riff.h"
#include "s2t_defs.h"

/* --------------------------------------------------------------- */
void DCRemove(SND_BUFFER signal, int count_of_samples){
/*	
	Functioning:	Remove the DC-component from signal.  
	Inputs:			SND_BUFFER signal		Signal array.
					int count_of_samples	count of samples in array.
					
	Outputs:		Signal without DC-component.

	Return value:	-
	History:		Teemu Kilpelinen 15.09.2000.
	Version:		1.0 (original)

	Reference:		
*/
	INT32 i;
	float sum = 0;
	float b,c;
    	
	for (i = 0; i < count_of_samples; i++)	    
		sum += (signal[i]/count_of_samples);
	
	for (i=0; i < count_of_samples; i++){
		b = signal[i];
        c = b - sum; 
		signal[i] = c;
		}
} /* DCRemove */


/* --------------------------------------------------------------- */
void HighEmphasisFilter(SND_BUFFER signal, int count_of_samples, 
	float high_emph_coeff){
/*
	Functioning:	Remove lowest frequencies from signal.  
	Inputs:			SND_BUFFER signal		Signal array.
					int count_of_samples	count of samples in array.
					float high_emph_coeff   high emphasis coefficient value.
						
	Outputs:		High emphasis filtered signal.

	Return value:	-
	History:		Teemu Kilpelinen 15.09.2000.
	Version:		1.0 (original)

	Reference:		
*/
	float b,c;
	SND_BUFFER signal_tmp;
	
    /* Memory allocation for tmp-buffer */
	signal_tmp = CreateNewFloatArray(count_of_samples);
			
	INT32 i;
	
	printf("High emphasis filtering...");

	for (i = 0; i < count_of_samples; i++)
        signal_tmp[i] = signal[i];
        	
	for (i = count_of_samples - 1; i >= 0; i--){
		
		if(i == 0){
			b = (high_emph_coeff * signal_tmp[count_of_samples - 1]);
			c = signal_tmp[i];
			}
	
		else{
			b = (high_emph_coeff * signal_tmp[i-1]);
			c = signal_tmp[i];
			}
			
		signal[i] =  c - b;
			
	}
	printf("OK\n");
	/* memory deallocation for temp array */
	deallocate(signal_tmp);
} /* HighEmphasisFilter */


/* --------------------------------------------------------------- */
void WindowCoefficients(FLOAT_ARRAY window_function, int window_type, 
	int window_length){
/*
	Functioning:	Compute coefficients for time domain window function.  
	
	Inputs:			FLOAT_ARRAY window_function    Values of window function 
					int window_type			       1 = Hamming, 2 = Barlett, 
											       3 = Hanning, 4 = Blackman
					int window_length              Window length in samples.
						
	Outputs:		Time domain window function.
	
	Return value:	-
	History:		Teemu Kilpelinen 15.04.2001.
	Version:		1.0 (original)

	Reference:		
*/
	int k;
	
	for(k = 0; k < window_length; k++){ 
        
		/* Hamming */
	    if (window_type == 1)   
			window_function[k] = (float)(0.54 - 0.46 * 
			cos(2 * PI * k /window_length));	
         
		/* Barlett (triangular) */
		else if (window_type == 2){
			if (k <= (window_length / 2.0))
				window_function[k] = (2.0f * k) / window_length;
			else window_function[k] = 2.0f - (2.0f * k)/window_length;
		}
		
		/* Hanning */
 		else if (window_type == 3) 
             window_function[k] = (float)(0.5 - 0.5 * 
			 cos(2 * PI * k / window_length));			

		/* Blackman */
		else 
			window_function[k] = (float)(0.42 - 0.5 * 
				cos(2 * PI * k / window_length) + 0.08 * 
				cos(4 * PI * k / window_length));
	}
} /* WindowCoefficients */


/* --------------------------------------------------------------- */
void PreProcess(FLOAT_ARRAY* window_function, SND_BUFFER signal, 
	FRAME_INFORMATION* frame_variables, SIGNAL_INFORMATION* signal_variables, 
	USER_PARAMETERS* user_params){
/*
	Functioning:	Make pre-process process  
	
	Inputs:			FLOAT_ARRAY window_function    Values of window function. 
					SND_BUFFER signal			   Signal.
					FRAME_INFORMATION frame_variables  Struct of frame variables.
					SIGNAL_INFORMATION	signal_variables Struct of signal 
										variables.
					USER_PARAMETERS user_params	   Struct of user parameters
								
	Outputs:		Pre-processed signal and window-function.
	
	Return value:	-
	History:		Teemu Kilpelinen 16.05.2001.
	Version:		1.0 (original)

	Reference:		
*/
	/* Calculates frame's length to samples. */
	frame_variables->samples_per_frame = 
		Ms2Samples(user_params->win_length, signal_variables->sampling_rate);
	
	/* Remove DC-offset. Used if wanted. */
	if (user_params->remove_dc_offset) 
		DCRemove(signal, signal_variables->count_of_samples);

	/* High emphasis filter. Used if wanted. */
	if (user_params->high_emph_coeff > 0)		
		HighEmphasisFilter(signal, signal_variables->count_of_samples, 
			user_params->high_emph_coeff);	

	/* Memory allocation for window_function */
	*window_function = 
		CreateNewFloatArray(frame_variables->samples_per_frame);
    
	/* Calculates values to window_fuction (depends of user choices: Hamming, 
	   Blackman, etc.). */
	WindowCoefficients(*window_function, user_params->win_function_type, 
		frame_variables->samples_per_frame);
}/* PreProcess */
