/*---------------------------------------------------------------------*/ 
/*  wav_io.cpp  [Don Cross] 16.2.1993, Teemu Kilpelinen 20.4.2001	   */
/*																	   */	
/*	This module includes functions for .wav -file handling.			   */																							
/*	Version Number: 0.01											   */
/*---------------------------------------------------------------------*/																							

#include <windows.h>
#include <stdio.h>
#include "helpfunctions.h"
#include "memctrl.h"
#include "riff.h"
#include "s2t_defs.h"
#include "wav_io.h"

#ifdef __cplusplus
extern "C"{
#endif

#include "interfc.h"

#ifdef __cplusplus
}	
#endif

#if defined(min) && defined(max)
#undef min
#undef max
#endif

/* --------------------------------------------------------------- */
int WaveReadTableSize(char* WaveFilename, int* BufferSize){
/*
  Functioning:  Count the size of wave signal.
	
  Inputs:		char* WaveFilename   Source wave signal file. 
				int* BufferSize      Size of buffer needed for wave.
  
  Outputs:		Size of buffer needed for wave.
	
  Return value:	error_code (int)
  History:		[Don Cross]
  Version:		1.0 (original)
  
Reference: -		
*/

	WaveFile Wave;
	DDCRET Value;
	//printf("Reading wave file %s", WaveFilename);
	
	if ((Value = Wave.OpenForRead(WaveFilename)) != DDC_SUCCESS) return Value;
	*BufferSize = Wave.NumChannels() * Wave.NumSamples(); 
	DDCRET Close();
	//printf("%d\n", *BufferSize);
	return DDC_SUCCESS;
}


/* --------------------------------------------------------------- */
int WaveReadContents(char* WaveFilename, INT16* Buffer, int BufferSize,
	int* Channels, int* SamplingRate, int* BitsPerSample, long* SamplesRead){
/*
  Functioning:  Read signal to INT16 -type buffer, clarify the signal 
				variables.
	
  Inputs:		char* WaveFilename   Return wave file.  
				INT16* Buffer		 Source buffer.
				int BufferSize,      Size of destination buffer. 
			    int* Channels		 Count of channels.
				int* SamplingRate    Sampling rate of signal.
				int* BitsPerSample	 Count of bits in signal.
				long* SamplesRead    Count of samples function managed to read.

  Outputs:		Wave signal in WaveFilename.
	
  Return value:	error_code (int)
  History:		[Don Cross]
  Version:		1.0 (original)
  
Reference: -		
*/
	long limit;

	WaveFile Wave;
	DDCRET Value;
	if ((Value = Wave.OpenForRead(WaveFilename)) != DDC_SUCCESS) return Value;
	
	printf("\n");
	*Channels = Wave.NumChannels();
	printf("Number of channels: %d\n", *Channels);
	
	*SamplingRate = Wave.SamplingRate();
	printf("Sampling rate: %d\n", *SamplingRate);
	
	*BitsPerSample = Wave.BitsPerSample();
	printf("Bits per sample: %d\n", *BitsPerSample); 

	printf("Reading .wav to buffer...\n");
    
	limit = ((unsigned)BufferSize < Wave.NumChannels() * Wave.NumSamples()) ? 
    BufferSize : (signed)(Wave.NumChannels() * Wave.NumSamples());

	for (*SamplesRead = 0; *SamplesRead < limit; (*SamplesRead)++){
		Value = Wave.ReadSample(&(Buffer[(*SamplesRead) * *Channels]));
	  if (Value != DDC_SUCCESS) return Value;
    }
	
	return DDC_SUCCESS;
} /* WaveReadContents */


/* --------------------------------------------------------------- */
int WaveWrite(char* WaveFilename, INT16* Buffer, int Samples,
		      int Channels, int SamplingRate, int BitsPerSample){
/*
  Functioning:  Wave write function.
	
  Inputs:		char* WaveFilename   Return wave file.  
				INT16* Buffer		 Source buffer.
				int Samples			 Count of samples.
		        int Channels		 Count of channels.
				int SamplingRate     Sampling rate of signal.
				int BitsPerSample	 Count of bits in signal.
	
  Outputs:		Wave signal in WaveFilename.
	
  Return value:	error_code (int)
  History:		[Don Cross]
  Version:		1.0 (original)
  
Reference: -		
*/
	WaveFile Wave;
	DDCRET Value = Wave.OpenForWrite(WaveFilename, SamplingRate,
		BitsPerSample, Channels);
	if (Value != DDC_SUCCESS) return Value;
	for (int k = 0; k < Samples; k++) {
		Value = Wave.WriteSample(&(Buffer[k * Channels]));
		if (Value != DDC_SUCCESS) return Value;
	}
	return DDC_SUCCESS;
} /* WaveWrite */


/* --------------------------------------------------------------- */
int ReadSignal(char* in_file, SND_BUFFER* signal, 
	SIGNAL_INFORMATION* signal_variables){
/*
  Functioning:  Read voice signal to buffer
	
  Inputs:		char* in_file							Input voice file. 
				SND_BUFFER signal						Input signal.
				SIGNAL_INFORMATION*  signal_variables   Struct of signal variables.
	
  Outputs:		voice signal (float array)
	
  Return value:	error_code (int)
  History:		Teemu Kilpelinen and Ismo Krkkinen 15.04.2001
  Version:		1.0 (original)
  
Reference:		
*/
	
	/* Calculates size of the input signal in samples */
	int error_code;
	INT16 *int_signal;
	
	error_code = WaveReadTableSize(in_file, &(signal_variables->
		wave_table_size));

	if (error_code != DDC_SUCCESS){
		ErrorMessage("Reading wave table size failed: %s\n", 
			DDCRET_String((DDCRET)error_code));
		
		return error_code;
	}
	
	/*Memory allocation for integer signal buffer */
	int_signal = (INT16*)(CreateNewIntArray(signal_variables->wave_table_size));
	
	/* Read signal to integer_buffer. At same time program finds out some 
	features of signal: sampling rate, bits per sample, etc.  */	
	error_code = WaveReadContents(in_file, int_signal, 
		signal_variables->wave_table_size, 
		&signal_variables->count_of_channels, 
		&signal_variables->sampling_rate, 
		&signal_variables->bits_per_sample, 
		&signal_variables->count_of_samples);
	
	if (error_code != DDC_SUCCESS){
		ErrorMessage("Reading wave file data failed: %s\n", 
			DDCRET_String((DDCRET)error_code));
		return error_code;
	};
	
	if (signal_variables->wave_table_size != 
		signal_variables->count_of_samples)
			printf("Reading error in input signal! Some samples cannot be read!");

	/*Memory allocation for signal */
	*signal = CreateNewFloatArray(signal_variables->count_of_samples);
	
	/*Copy int-type buffer to float-type buffer */
	error_code = Int16ToFloat(int_signal, *signal, 
		signal_variables->count_of_samples);
	
	/*Memory deallocation for integer_buffer */
	deallocate(int_signal);
    
	return error_code;
} /* ReadSignal */
